@@ -14,7 +14,8 @@ from roborock.data import (
     HomeDataProduct,
     UserData,
 )
-from roborock.devices.device import RoborockDevice
+from roborock.devices.device import DeviceReadyCallback, RoborockDevice
+from roborock.exceptions import RoborockException
 from roborock.map.map_parser import MapParserConfig
 from roborock.mqtt.roborock_session import create_lazy_mqtt_session
 from roborock.mqtt.session import MqttSession
@@ -68,22 +69,28 @@ class DeviceManager:
         self._devices: dict[str, RoborockDevice] = {}
         self._mqtt_session = mqtt_session
 
-    async def discover_devices(self) -> list[RoborockDevice]:
+    async def discover_devices(self, prefer_cache: bool = True) -> list[RoborockDevice]:
         """Discover all devices for the logged-in user."""
         cache_data = await self._cache.get()
-        if not cache_data.home_data:
-            _LOGGER.debug("No cached home data found, fetching from API")
-            cache_data.home_data = await self._web_api.get_home_data()
+        if not cache_data.home_data or not prefer_cache:
+            _LOGGER.debug("Fetching home data (prefer_cache=%s)", prefer_cache)
+            try:
+                cache_data.home_data = await self._web_api.get_home_data()
+            except RoborockException as ex:
+                if not cache_data.home_data:
+                    raise
+                _LOGGER.debug("Failed to fetch home data, using cached data: %s", ex)
             await self._cache.set(cache_data)
         home_data = cache_data.home_data
 
         device_products = home_data.device_products
-        _LOGGER.debug("Discovered %d devices %s", len(device_products), home_data)
+        _LOGGER.debug("Discovered %d devices", len(device_products))
 
         # These are connected serially to avoid overwhelming the MQTT broker
         new_devices = {}
         start_tasks = []
         for duid, (device, product) in device_products.items():
+            _LOGGER.debug("[%s] Discovered device %s %s", duid, product.summary_info(), device.summary_info())
             if duid in self._devices:
                 continue
             new_device = self._device_creator(home_data, device, product)
@@ -155,6 +162,7 @@ async def create_device_manager(
     cache: Cache | None = None,
     map_parser_config: MapParserConfig | None = None,
     session: aiohttp.ClientSession | None = None,
+    ready_callback: DeviceReadyCallback | None = None,
 ) -> DeviceManager:
     """Convenience function to create and initialize a DeviceManager.
 
@@ -163,6 +171,7 @@ async def create_device_manager(
         cache: Optional cache implementation to use for caching device data.
         map_parser_config: Optional configuration for parsing maps.
         session: Optional aiohttp ClientSession to use for HTTP requests.
+        ready_callback: Optional callback to be notified when a device is ready.
 
     Returns:
         An initialized DeviceManager with discovered devices.
@@ -211,7 +220,11 @@ async def create_device_manager(
                     raise NotImplementedError(f"Device {device.name} has unsupported B01 model: {product.model}")
             case _:
                 raise NotImplementedError(f"Device {device.name} has unsupported version {device.pv}")
-        return RoborockDevice(device, product, channel, trait)
+
+        dev = RoborockDevice(device, product, channel, trait)
+        if ready_callback:
+            dev.add_ready_callback(ready_callback)
+        return dev
 
     manager = DeviceManager(web_api, device_creator, mqtt_session=mqtt_session, cache=cache)
     await manager.discover_devices()
